/*  glimGls.cc

    Mark Woolrich, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

#include <strstream>

#include "glimGls.h"
#include "miscmaths.h"
#include "Log.h"
#include "Volume.h"
#include "VolumeSeries.h"

#ifndef NO_NAMESPACE
using namespace MISCMATHS;
using namespace UTILS;
using namespace TACO;
namespace SIGPROC {
#endif

  void GlimGls::setData(const ColumnVector& y, const Matrix& x, const int ind)
    {
      // compute b
      Matrix inv_xx = (x.t()*x).i();
      b.Column(ind) = inv_xx*x.t()*y;

      // compute r
      Matrix I(sizeTS, sizeTS);
      Identity(I);
      ColumnVector r = (I-x*inv_xx*x.t())*y;

      // compute sigma squareds 
      sigmaSquareds(ind) = (r.t()*r/sizeTS).AsScalar();

      // set corrections
      SetCorrection(inv_xx, ind);
    }

  void GlimGls::Save(const VolumeSeries::Dims& pdims, const ColumnVector& prethreshpos)
    {
      // Need to save b, sigmaSquareds, corrections and dof 
      Log& logger = Log::getInstance();
     
      // b:
      Volume peVol;
      for(int i = 1; i <= numParams; i++)
	{
	  peVol = b.Row(i).AsColumn();
	  peVol.setDims(pdims);
	  peVol.setPreThresholdPositions(prethreshpos);
	  peVol.unthreshold();	
	  
	  // Add param number to "pe" to create filename:
	  char strc[4];
	  ostrstream osc(strc,4);
	  osc << i << '\0';
	  
	  peVol.writeAsFloat(logger.getDir() + "/pe" + strc);
	}

      // sigmaSquareds:
      sigmaSquareds.setDims(pdims);
      sigmaSquareds.setPreThresholdPositions(prethreshpos);
      sigmaSquareds.unthreshold();	
      sigmaSquareds.writeAsFloat(logger.getDir() + "/sigmasquareds");

      // dof:
      ColumnVector dofVec(1);
      dofVec = dof;
      logger.out("dof", dofVec);

      // corrections:
      logger.out("corrections", corrections);
    }

  void GlimGls::SetCorrection(const Matrix& corr, const int ind)
    {
      Tracer ts("SetCorrection");

      // puts Matrix corr which is p*p into Matrix correction
      // as a p*p length row:
      int p = corr.Nrows();
      
      for (int i = 1; i <= p; i++)
	{
	  for (int j = 1; j <= p; j++)
	    {
	      corrections(ind, (i-1)*p + j) = corr(i,j); 
	    }
	}
    }

#ifndef NO_NAMESPACE
}
#endif







