/*  generalio.cc

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

// General IO functions (images and transformation files)


#include <cstdlib>
#include <iostream>
#include <fstream>
#include <strstream>
#include "avwio.h"
#include "generalio.h"
#include "miscmaths.h"
#include "miscimfns.h"

#ifndef NO_NAMESPACE
 using namespace MISCMATHS;
 using namespace MISCIMFNS;
 using namespace MJIMAGE;

 namespace GENERALIO {
#endif


float* AvwReadBasic(AVW* IP, bool read_img_data=true)
{
  short x,y,z,v,type;
  AvwGetDim(IP,&x,&y,&z,&v);
  size_t imagesize=x*y*z*v;
  AvwGetDataType(IP,&type);

  float* fbuffer=new float[imagesize];
  
  if (read_img_data) {
    switch(type)
      {
      case DT_SIGNED_SHORT:
	{
	  short* sbuffer=new short[imagesize];
	  AvwReadVolumes(IP,sbuffer,v);
	  for(size_t i=0; i<imagesize; i++) fbuffer[i]=(short) sbuffer[i];
	  delete[] sbuffer;
	}
	break;
      case DT_UNSIGNED_CHAR:
	{
	  unsigned char* sbuffer=new unsigned char[imagesize];
	  AvwReadVolumes(IP,sbuffer,v);
	  for(size_t i=0; i<imagesize; i++) 
	    { fbuffer[i]=(unsigned char) sbuffer[i]; }
	  delete[] sbuffer;
	}
	break;
      case DT_SIGNED_INT:
	{
	  int* sbuffer=new int[imagesize];
	  AvwReadVolumes(IP,sbuffer,v);
	  for(size_t i=0; i<imagesize; i++) fbuffer[i]=(int) sbuffer[i];
	  delete[] sbuffer;
	}
	break;
      case DT_DOUBLE:
	{
	  double* sbuffer=new double[imagesize];
	  AvwReadVolumes(IP,sbuffer,v);
	  for(size_t i=0; i<imagesize; i++) fbuffer[i]=(double) sbuffer[i];
	  delete[] sbuffer;
	}
	break;
      case DT_FLOAT:
	AvwReadVolumes(IP,fbuffer,v);
	break;
      default:
	ostrstream errmsg;
	errmsg << "Avwread: DT " << type <<  " not supported";
	perror(errmsg.str());
      }
  }
  return fbuffer;
}


int AvwWriteVolumes(AVW* OP, const float *iptr, const size_t no_vols, 
		    const short AVW_DATA_TYPE)
{
  // assumes the the dimensions of the AVW are already set!
  AvwSetDataType(OP, AVW_DATA_TYPE);
  size_t imagesize=AvwGetVolSize(OP) * no_vols;

  switch(AVW_DATA_TYPE)
    {
    case DT_SIGNED_SHORT:
      {
	short* sbuffer=new short[imagesize];
	for(size_t i=0; i<imagesize; i++) sbuffer[i]=(short) *(iptr++);
	AvwWriteVolumes(OP,sbuffer,no_vols);
	delete[] sbuffer;
      }
      break;
    case DT_UNSIGNED_CHAR:
      {
	unsigned char* sbuffer=new unsigned char[imagesize];
	for(size_t i=0; i<imagesize; i++) sbuffer[i]=(unsigned char) *(iptr++);
	AvwWriteVolumes(OP,sbuffer,no_vols);
	delete[] sbuffer;
      }
      break;
    case DT_SIGNED_INT:
      {
	int* sbuffer=new int[imagesize];
	for(size_t i=0; i<imagesize; i++) sbuffer[i]=(int) *(iptr++);
	AvwWriteVolumes(OP,sbuffer,no_vols);
	delete[] sbuffer;
      }
      break;
    case DT_DOUBLE:
      {
	double* sbuffer=new double[imagesize];
	for(size_t i=0; i<imagesize; i++) sbuffer[i]=(double) *(iptr++);
	AvwWriteVolumes(OP,sbuffer,no_vols);
	delete[] sbuffer;
      }
      break;
    case DT_FLOAT:
      {
	AvwWriteVolumes(OP,iptr,no_vols);
      }
      break;
    default:
      ostrstream errmsg;
      errmsg << "AvwWriteVolumes: DT " << AVW_DATA_TYPE <<  " not supported";
      perror(errmsg.str());
    }
  return 0;
}



void AvwWrite(AVW* OP, const volume& outputvol, const short dtype)
{
  AvwSetDim(OP,outputvol.xsize(),outputvol.ysize(),outputvol.zsize(),1);
  AvwSetDataType(OP, dtype);
  AvwWriteVolumes(OP,&(outputvol(0,0,0)),1,dtype);
}

void AvwWrite(AVW* OP, const volume& outputvol)
{
  AvwWrite(OP,outputvol,DT_FLOAT);
}


void AvwRead(AVW* IP, volume& output, bool read_img_data=true) {
  short x,y,z,v;
  AvwGetDim(IP,&x,&y,&z,&v);
  float* buf = AvwReadBasic(IP,read_img_data);
  
  output.reinitialize(x,y,z,buf);
  
  return;
}



////////////////////////////////////////////////////////////////////////////

// VOLUME I/O

int make_basename(string& filename)
{
  Tracer tr("make_basename");
  if ( filename.size()<1 ) return -1;
  string basename = filename;
  int fsize = basename.length(), indx;
  // find either .hdr or .img extensions and remove them
  indx=basename.rfind(".hdr");
  if ((indx>=0) && (indx<fsize)) {
    basename.erase(indx,fsize);
    fsize = basename.length();
  }
  indx=basename.rfind(".img");
  if ((indx>=0) && (indx<fsize)) {
    basename.erase(indx,fsize);
    fsize = basename.length();
  }
  indx = basename.length()-1;
  if (basename[indx] == '.') {
    basename.erase(indx,fsize);
    fsize = basename.length();
  }
  // this is now the basename
  filename = basename;
  return 0;
}


void get_avw_origin(volume& vol, AVW* IP1) {
  short orig[5];
  AvwGetOriginator(IP1,orig);
  ColumnVector origin(3);
  origin(1) = orig[0];
  origin(2) = orig[1];
  origin(3) = orig[2];
  
  vol.avw_origin = origin;
}


int read_volume(volume& target, const string& filename, short& dtype,
		bool read_img_data)
{
  Tracer trcr("read_volume");
  if ( filename.size()<1 ) return -1;
  string basename = filename, fname;
  make_basename(basename);

  // check that the .hdr file exists
  fname = basename + ".hdr";
  ifstream ftest1(fname.c_str());
  if (!ftest1) {
    cerr << "Cannot open volume " << fname << " for reading!\n";
    exit(1);
  }
  ftest1.close();

  // check that the .img file exists
  fname = basename + ".img";
  ifstream ftest2(fname.c_str());
  if (!ftest2) {
    cerr << "Cannot open volume " << fname << " for reading!\n";
    exit(1);
  }
  ftest2.close();
  
  AVW* IP1 = AvwOpen(basename.c_str(), "r");
  AvwRead(IP1,target,read_img_data); 

  float x,y,z,tr;
  AvwGetVoxDim(IP1,&x,&y,&z,&tr);
  if (x<0.0) x=-x;
  if (y<0.0) y=-y;
  if (z<0.0) z=-z;
  target.setvoxelsize(x,y,z);

  AvwGetDataType(IP1,&dtype);

  get_avw_origin(target,IP1);

  AvwClose(IP1);
  return 0;
}


int read_volume(volume& target, const string& filename)
{
  short dtype;
  int retval = read_volume(target,filename,dtype,true);
  return retval;
}

int read_volume_hdr_only(volume& target, const string& filename)
{
  short dtype;
  int retval = read_volume(target,filename,dtype,false);
  return retval;
}

void set_avw_origin(const volume& vol, AVW* IP1) {
  short orig[5];
  if (vol.avw_origin.Nrows()<3) {
      vol.avw_origin.ReSize(3);  
      vol.avw_origin = 0;  
  }
  orig[0] = (short) vol.avw_origin(1);
  orig[1] = (short) vol.avw_origin(2);
  orig[2] = (short) vol.avw_origin(3);
  orig[3] = 0;
  orig[4] = 0;

  AvwSetOriginator(IP1,orig);
}

int save_volume(const volume& source, const string& filename, const short dtype)
{
  Tracer tr("save_volume");
  string basename = filename;
  make_basename(basename);
  if ( basename.size()<1 ) return -1;

  AVW* OP=AvwOpen(basename.c_str(),"w");
  if (OP==0) return -1;
  AvwSetVoxDim(OP,source.getx(),source.gety(),source.getz(),1.0);
    
  float fmin, fmax;
  get_min_max(source,fmin,fmax);
  int imin, imax;
  imin = (int) fmin;
  imax = (int) fmax;
  if (imax<=imin)  imax=imin+1;
  AvwSetMinMax(OP,imin,imax);
  set_avw_origin(source,OP);
    
  AvwWrite(OP,source,dtype);
  AvwClose(OP); 
  return 0;
}


int save_volume(const volume& source, const string& filename)
{
  return save_volume(source,filename,DT_FLOAT);
}

  
int read_series(tvolume& time_series, const string& filename, short& dtype)
{
  Tracer trcr("read_series");
  AVW *avw;
  short X, Y, Z, V;
  string basename = filename, fname;
  float x,y,z,tr;

  if ( filename.size()<1 ) return -1;
  make_basename(basename);
  
  // check that the .hdr file exists
  fname = basename + ".hdr";
  ifstream ftest1(fname.c_str());
  if (!ftest1) {
    cerr << "Cannot open volume " << fname << " for reading!\n";
    exit(1);
  }
  ftest1.close();

  // check that the .img file exists
  fname = basename + ".img";
  ifstream ftest2(fname.c_str());
  if (!ftest2) {
    cerr << "Cannot open volume " << fname << " for reading!\n";
    exit(1);
  }
  ftest2.close();

  avw = AvwOpen(basename.c_str(), "r");
  AvwGetDim(avw, &X, &Y, &Z, &V); 

  AvwGetVoxDim(avw,&x,&y,&z,&tr);
  if (x<0.0) x=-x;
  if (y<0.0) y=-y;
  if (z<0.0) z=-z;

  time_series.reinitialize(V, tr);

  float* buffer, *volbuffer;
  long int volsize;

  volsize = X*Y*Z;

  buffer = AvwReadBasic(avw); 
  volbuffer = buffer;

  for (int i = 0; i < V; i++) {
    time_series[i].reinitialize(X, Y, Z, volbuffer);
    time_series[i].setvoxelsize(x,y,z);
    get_avw_origin(time_series[i],avw);
    volbuffer += volsize;
  }

  AvwGetDataType(avw,&dtype);
  
  AvwClose(avw);
  return 0;
}

int read_series(tvolume& time_series, const string& filename)
{
  short dtype;
  int retval = read_series(time_series,filename,dtype);
  return retval;
}


int save_series(const tvolume& time_series, const string& filename, 
		const short dtype)
{
  Tracer tr("save_series");
  if ( filename.size()<1 ) return -1;
  AVW* avw=AvwOpen(filename.c_str(),"w");
  if (avw==0) return -1;
  AvwSetVoxDim(avw,time_series[0].getx(),time_series[0].gety(), time_series[0].getz(), time_series.getTR());
    
  float fmin, fmax;
  float oldmin, oldmax;
  
  get_min_max(time_series[0],oldmin,oldmax);
  
  if (time_series.getSize() > 1) {
    for (int i=1; i<time_series.getSize(); i++) {
      get_min_max(time_series[i],fmin,fmax);
      if (oldmax < fmax) oldmax = fmax;
      if (oldmin > fmin) oldmin = fmin;
    }
  }
  fmin = oldmin; fmax = oldmax;
  
  AvwSetMinMax(avw,(int) fmin,(int) (fmax+0.9999));
  set_avw_origin(time_series[0],avw);

  AvwSetDim(avw,time_series[0].xsize(),time_series[0].ysize(), time_series[0].zsize(), time_series.getSize());
  

  AvwSetDataType(avw, dtype);
  
  for (int t=0; t<time_series.getSize(); t++) 
    AvwWriteVolumes(avw,&(time_series[t](0,0,0)),1,dtype);

  AvwClose(avw); 
  return 0;
}


int save_series(const tvolume& time_series, const string& filename)
{
  return save_series(time_series,filename,DT_FLOAT);
}

//////////////////////////////////////////////////////////////////////////

// MATRIX I/O


int read_ascii_matrix(Matrix &target, ifstream& matfile)
{
  Tracer tr("read_ascii_matrix");
  for (int i=1; i<=target.Nrows(); i++) {
    for (int j=1; j<=target.Ncols(); j++) {
      matfile >> target(i,j);
    }
  }
  return 0;
}


int read_ascii_matrix(Matrix &target, const string& filename)
{
  Tracer tr("read_ascii_matrix");
  if ( filename.size()<1 ) return -1;
  ifstream matfile(filename.c_str());
  if (!matfile) { 
    cerr << "Could not open matrix file " << filename << endl;
    return -1;
  }
  int rval = read_ascii_matrix(target,matfile);
  matfile.close();
  return rval;
}


int get_medx_small_matrix(Matrix &target, ifstream& matfile)
{
  Tracer tr("get_medx_small_matrix");
  string str1;
  matfile >> str1;
  if (str1 != "[") {
    return -1;
  }
  int i=1, j=1;
  matfile >> str1;
  while (str1 != "]") {
    target(i,j) = atof(str1.c_str());
    if (j==4) { i++; j=1; }
    else { j++; }
    matfile >> str1;
  }
  return 0;
}  


int get_medx_matrix(Matrix &target, ifstream& matfile)
{
  Tracer tr("get_medx_matrix");
  string str1, str2;
  matfile >> str1 >> str2;
  if ((str1 != "<<") || (str2 != "/matrix")) {
    return -1;
  }
  target.ReSize(4,4);
  Identity(target);  // default return value
  return get_medx_small_matrix(target,matfile);
}  


int get_minc_matrix(Matrix &target, ifstream& matfile)
{
  Tracer tr("get_minc_matrix");
  string str1;
  matfile >> str1;
  if (str1 != "=") {
    cerr << "Could not parse MINC transform file" << endl;
    return -1;
  }
  target.ReSize(4,4);
  Identity(target);  // default return value
  for (int i=1; i<=3; i++) {
    for (int j=1; j<=4; j++) {
      matfile >> str1;
      target(i,j) = atof(str1.c_str());
    }
  }
  return 0;
}  


int medx2world(Matrix &target, const Matrix& medxmat, 
	       const ColumnVector& finalimsize, const Matrix& finalvoxsize,
	       const volume& initvol) 
{
  Tracer tr("medx2world");
  // convert the voxel->voxel MEDx form to a world->world form
  Matrix swapy1(4,4), samp1(4,4);
  Matrix swapy2(4,4), samp2(4,4);
  Identity(swapy1);
  swapy1(2,2)=-1;
  swapy2=swapy1;
  swapy1(2,4)=initvol.ysize()-1.0;  // corrected version
  swapy2(2,4)=finalimsize(2)-1.0;   // corrected version
  Identity(samp1);
  samp1(1,1) = initvol.getx();  
  samp1(2,2) = initvol.gety();
  samp1(3,3) = initvol.getz();
  samp2 = finalvoxsize;
  samp2(1,4)=0; samp2(2,4)=0; samp2(3,4)=0;  // translation can be ignored
  samp2(1,1)=fabs(samp2(1,1));    // swapping info is not used
  samp2(2,2)=fabs(samp2(2,2)); 
  samp2(3,3)=fabs(samp2(3,3));
  target = samp2 * swapy2 * medxmat * swapy1 * samp1.i(); // correct
  return 0;
}


int world2medx(Matrix &medxmat, const Matrix& worldmat, 
	       const ColumnVector& finalimsize, const Matrix& finalvoxsize,
	       const volume& initvol) 
{
  Tracer tr("world2medx");
  // convert the world->world form to a voxel->voxel MEDx form 
  Matrix swapy1(4,4), samp1(4,4);
  Matrix swapy2(4,4), samp2(4,4);
  Identity(swapy1);
  swapy1(2,2)=-1;
  swapy2=swapy1;
  swapy1(2,4)=initvol.ysize()-1.0;  // corrected version
  swapy2(2,4)=finalimsize(2)-1.0;   // corrected version
  Identity(samp1);
  samp1(1,1) = initvol.getx();  
  samp1(2,2) = initvol.gety();
  samp1(3,3) = initvol.getz();
  samp2 = finalvoxsize;
  samp2(1,4)=0; samp2(2,4)=0; samp2(3,4)=0;  // translation can be ignored
  samp2(1,1)=fabs(samp2(1,1));    // swapping info is not used
  samp2(2,2)=fabs(samp2(2,2)); 
  samp2(3,3)=fabs(samp2(3,3));
  medxmat = swapy2 * samp2.i() * worldmat * samp1 * swapy1;
  return 0;
}


int minc2world(Matrix &target, const Matrix& mincmat, 
	       const volume& initvol, const volume& finalvol)
{
  Tracer tr("minc2world");
  // convert the voxel->voxel MINC form to a world->world form
  Matrix trans1(4,4), trans2(4,4), samp1(4,4), samp2(4,4), 
         flipx1(4,4), flipx2(4,4);

  Identity(trans1);
  trans1.SubMatrix(1,3,4,4) = initvol.avw_origin;

  Identity(trans2);
  trans2.SubMatrix(1,3,4,4) = finalvol.avw_origin;

  Identity(samp1);
  samp1(1,1) = fabs(initvol.getx());  
  samp1(2,2) = fabs(initvol.gety());
  samp1(3,3) = fabs(initvol.getz());

  Identity(samp2);
  samp2(1,1) = fabs(finalvol.getx());  
  samp2(2,2) = fabs(finalvol.gety());
  samp2(3,3) = fabs(finalvol.getz());
  
  Identity(flipx1);
  flipx1(1,1) = -1.0;
  flipx1(1,4) = initvol.xsize() - 1.0;

  Identity(flipx2);
  flipx2(1,1) = -1.0;
  flipx2(1,4) = finalvol.xsize() - 1.0;

  target = samp2 * flipx2 * trans2 * samp2.i() * mincmat * samp1 *
                  trans1.i() * flipx1 * samp1.i();
  return 0;
}



int read_medx_matrix(Matrix &target, ifstream& matfile, const volume& invol)
{
  Tracer tr("read_medx_matrix");
  string curword;
  Matrix id4(4,4), gmat, imsize(1,3), voxsize(4,4), medxmat(4,4);
  Identity(id4);  // set to be the identity matrix for checking
  medxmat = id4;   // default return value
  imsize = 1.0;
  voxsize = id4;
  while (!matfile.eof()) {
    matfile >> curword;
    if ( (curword == "/outputsize") ) {
      get_medx_small_matrix(imsize,matfile);
    }
    if ( (curword == "/outputusermatrix") ) {
      get_medx_small_matrix(voxsize,matfile);
    }
    if ( (curword == "/GenericReslice")  || (curword == "/ShadowTransform")
	 || (curword == "/MotionCorrectionReslice") 
	 || (curword == "/UserTransformation") 
	 || (curword == "/IntoTalairachSpace") ) {
      if (get_medx_matrix(gmat,matfile)==0) {  // only accept good returns
	if (SumSquare(gmat - id4) > 0.001) // only accept non-identities
	  medxmat = gmat;
      }
    }
  }
  ColumnVector imsizev(3);
  imsizev(1) = imsize(1,1);
  imsizev(2) = imsize(1,2);
  imsizev(3) = imsize(1,3);
  medx2world(target,medxmat,imsizev,voxsize,invol);
  return 0;
}


int read_minc_matrix(Matrix &target, ifstream& matfile, const volume& invol,
		     const volume& finalvol)
{
  Tracer tr("read_minc_matrix");
  string curword;
  Matrix gmat, id4(4,4);
  Identity(id4);
  target = id4;  // default return value
  while (!matfile.eof()) {
    matfile >> curword;
    if ( (curword == "Linear_Transform") ) {
      get_minc_matrix(gmat,matfile);
    }
    curword="";
  }
  minc2world(target,gmat,invol,finalvol);
  return 0;
}


int read_matrix(Matrix &target, const string& filename, const volume& invol,
		const volume& finalvol)
{
  Tracer tr("read_matrix");
  if ( filename.size()<1 ) return -1;
  ifstream matfile(filename.c_str());
  if (!matfile) { 
    cerr << "Could not open matrix file " << filename << endl;
    return -1;
  }
  string firstline;
  matfile >> firstline;
  matfile.seekg(0,ios::beg);
  int rval=0;
  if (firstline == "%!VEST-Transformations")
    {
      rval = read_medx_matrix(target,matfile,invol);
    } else if (firstline == "MNI") {
      if (finalvol.xsize()>0) {
	rval = read_minc_matrix(target,matfile,invol,finalvol);
      } else {
	cerr << "Cannot read MINC transform files with finalvol" << endl;
	rval = -1;
      }
    } else {
      rval = read_ascii_matrix(target,matfile);
    }
  matfile.close();
  return rval;
}


int read_matrix(Matrix &target, const string& filename, const volume& invol)
{
  volume dummy(0,0,0);
  return read_matrix(target,filename,invol,dummy);
}

//------------------------------------------------------------------------//

int write_ascii_matrix(const Matrix& affmat, const string& filename)
{
  Tracer tr("write_ascii_matrix");
  if ( (filename.size()<1) ) return -1;
  ofstream matfile(filename.c_str());
  if (!matfile) { 
    cerr << "Could not open file " << filename << " for writing" << endl;
    return -1;
  }

  for (int i=1; i<=affmat.Nrows(); i++) {
    for (int j=1; j<=affmat.Ncols(); j++) {
      matfile << affmat(i,j) << "  ";
    }
    matfile << endl;
  }
  matfile.close();
  return 0;
}



int put_medx_matrix(ofstream& matfile, const string& name, const Matrix& affmat)
{
  Tracer tr("put_medx_matrix");
  if (affmat.Nrows()<=0) { return -1; }
  matfile << "        /" << name << " [" << endl;
  for (int i=1; i<=affmat.Nrows(); i++) {
    for (int j=1; j<=affmat.Ncols(); j++) {
      matfile << "            " << affmat(i,j) << endl;
    }
  }
  matfile << "        ]" << endl;
  return 0;
}


int get_outputusermat(const string& filename, Matrix& oumat)
{
  Tracer trcr("get_outputusermat");
  if ( filename.size()<1 ) return -1;
  string basename = filename, fname;
  make_basename(basename);

  // check that the .hdr file exists
  fname = basename + ".hdr";
  ifstream ftest1(fname.c_str());
  if (!ftest1) {
    cerr << "Cannot open volume " << fname << " for reading!\n";
    exit(1);
  }
  ftest1.close();

  AVW* IP1(AvwOpen(basename.c_str(), "r"));
  short orig[5];
  AvwGetOriginator(IP1,orig);
  ColumnVector origin(3);
  origin(1) = orig[0];
  origin(2) = orig[1];
  origin(3) = orig[2];
  //cerr << "Origin (pre-swap) = " << origin.t() << endl;

  float x,y,z,tr;
  AvwGetVoxDim(IP1,&x,&y,&z,&tr);
  if (origin.MaximumAbsoluteValue() > 1e-5) {
    y = -fabs(y);  // MEDx insists that the y-voxel dim will be negative!!
  }
  //cerr << "Voxel dims are: " << x << " , " << y << " , " << z << endl;
  // Convert the SPM origin (as stored in the .hdr) to MEDx origin
  origin = 1.0 - origin;
  // Now deal with any negative voxel dimensions 
  //  - this behaviour is UNTESTED!  I assume that AIR behaves like this
  //    it is certainly true for the negative y-voxel induced by an SPM-origin
  short sx,sy,sz,st;
  AvwGetDim(IP1,&sx,&sy,&sz,&st);
  if (x<0.0) origin(1) += sx;
  if (y<0.0) origin(2) += sy;
  if (z<0.0) origin(3) += sz;
  //cerr << "Origin (post-swap) = " << origin.t() << endl;

  oumat.ReSize(4,4);
  Identity(oumat);
  oumat(1,1) = x;
  oumat(2,2) = y;
  oumat(3,3) = z;
  oumat(1,4) = origin(1);
  oumat(2,4) = origin(2);
  oumat(3,4) = origin(3);

  AvwClose(IP1);
  return 0;
}  


int write_medx_matrix(const Matrix& worldmat, const string& filename, 
		      const volume& initvol, 
		      const volume& finalvol, 
		      const string& mtype, const string& reffname)
{
  Tracer tr("write_medx_matrix");
  if ( (filename.size()<1) ) return -1;
  ofstream matfile(filename.c_str());
  if (!matfile) { 
    cerr << "Could not open file " << filename << " for writing" << endl;
    return -1;
  }

  // convert from world->world transform to the voxel->voxel MEDx transform
  Matrix swapy1(4,4), swapy2(4,4), samp1(4,4), samp2(4,4), medxtrans(4,4);
  Identity(samp1);  Identity(samp2);  Identity(swapy1);  Identity(swapy2);
  swapy1(2,2) = -1.0;
  swapy2(2,2) = -1.0;
  swapy1(2,4) = initvol.rows()-1.0;  // corrected version
  swapy2(2,4) = finalvol.rows()-1.0; // corrected version
  samp1(1,1) = initvol.getx();  
  samp1(2,2) = initvol.gety();
  samp1(3,3) = initvol.getz();
  samp2(1,1) = finalvol.getx();  
  samp2(2,2) = finalvol.gety();
  samp2(3,3) = finalvol.getz();
  medxtrans = swapy2 * samp2.i() * worldmat * samp1 * swapy1;


  string mattype;
  if ( (mtype.size()<1) || // default
       (mtype[0] == 'g') || (mtype[0] == 'G') )  
    { mattype="GenericReslice"; }
  if ( (mtype[0] == 's') || (mtype[0] == 'S') )  
    { mattype="ShadowTransform"; }
  if ( (mtype[0] == 't') || (mtype[0] == 'T') )  
    { mattype="IntoTalairachSpace"; }
  if ( (mtype[0] == 'u') || (mtype[0] == 'U') )  
    { mattype="UserTransformation"; }
  if ( (mtype[0] == 'm') || (mtype[0] == 'M') || 
       (mtype[0] == 'a') || (mtype[0] == 'A') )
    { mattype="MotionCorrectionReslice"; }

  matfile << "%!VEST-Transformations" << endl;
  matfile << "<<" << endl;
  matfile << "    /" << mattype << "     <<" << endl;
  put_medx_matrix(matfile,"matrix",medxtrans);
  matfile << "        /order  1" << endl;

  Matrix imsize(1,3);
  imsize(1,1) = finalvol.columns();
  imsize(1,2) = finalvol.rows();
  imsize(1,3) = finalvol.slices();
  put_medx_matrix(matfile,"outputsize",imsize);

  Matrix oumat(4,4);
  get_outputusermat(reffname,oumat);
  put_medx_matrix(matfile,"outputusermatrix",oumat);

  if (mattype == "MotionCorrectionReslice") {
    matfile << "        /outputunits    (mm)" << endl;
    matfile << "        /talairachcalibrated?   false" << endl;
  }
  if (mattype == "IntoTalairachSpace") {
    matfile << "        /talairachcalibrated?   true" << endl;
  }
  matfile << "    >>" << endl;
  matfile << ">>" << endl;
  matfile.close();
  return 0;
}


#ifndef NO_NAMESPACE
 }
#endif









