/*  interpolation.cc

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

// Interpolation functions


#include "interpolation.h"
#include <iostream>
#include <math.h>

#ifndef NO_NAMESPACE
 namespace INTERPOLATION {
#endif

  // INTERPOLATION STUFF

  float basic_tri_interpolation(const volume& v, 
				const float dx, const float dy, const float dz,
				const float v000, const float v001,
				const float v010, const float v011,
				const float v100, const float v101,
				const float v110, const float v111)
    {
      const float dx1=1.0-dx, dy1=1.0-dy, dz1=1.0-dz;
 

      return ( dx * ( dy * ( v111*dz  + 
			     v110*dz1 ) +
		      dy1 * ( v101*dz + 
			      v100*dz1 ) ) + 
	       dx1 * ( dy * ( v011*dz +
			      v010*dz1 )  +
		       dy1 * ( v001*dz  + 
			       v000*dz1 ) ) );
			   
    }


//    float q_tri_interpolation(const volume& v, 
//  			    const float x, const float y, const float z,
//  			    const int ix, const int iy, const int iz)
//      {
//        const float dx=x-ix, dy=y-iy, dz=z-iz;
//        const float dx1=1.0-dx, dy1=1.0-dy, dz1=1.0-dz;
 

//        return ( dx * ( dy * ( v(ix+1,iy+1,iz+1)*dz  + 
//  			     v(ix+1,iy+1,iz)*dz1 ) +
//  		      dy1 * ( v(ix+1,iy,iz+1)*dz + 
//  			      v(ix+1,iy,iz)*dz1 ) ) + 
//  	       dx1 * ( dy * ( v(ix,iy+1,iz+1)*dz +
//  			      v(ix,iy+1,iz)*dz1 )  +
//  		       dy1 * ( v(ix,iy,iz+1)*dz  + 
//  			       v(ix,iy,iz)*dz1 ) ) );
			   
//      }

  
  float tri_interpolation(const volume& v, const float x, const float y,
			  const float z)
    {
      int ix0, iy0, iz0, ix1, iy1, iz1;
      ix0 = (int) floor(x);
      iy0 = (int) floor(y);
      iz0 = (int) floor(z);
      ix1 = ix0 + 1;
      iy1 = iy0 + 1;
      iz1 = iz0 + 1;
      if ((ix0==v.columns()-1) && ((x-(float)ix0)<1e-5))
	{
	  ix1=ix0;
	}
      if ((iy0==v.rows()-1) && ((y-(float)iy0)<1e-5))
	{
	  iy1=iy0;
	}
      if ((iz0==v.slices()-1) && ((z-(float)iz0)<1e-5))
	{
	  iz1=iz0;
	}
      if ( (!v.in_bounds(ix0,iy0,iz0)) || (!v.in_bounds(ix1,iy1,iz1)) )
	{
	  cerr << "Tri_interpolation: Out of bounds!\n";
	  cerr << "(" << x << "," << y << "," << z << ") vs ";
	  cerr << "(" << v.columns() << "," << v.rows() << "," << 
	    v.slices() << ")" << endl;
	}
      return basic_tri_interpolation(v,x-ix0,y-iy0,z-iz0,v(ix0,iy0,iz0),
				     v(ix0,iy0,iz1),v(ix0,iy1,iz0),
				     v(ix0,iy1,iz1),v(ix1,iy0,iz0),
				     v(ix1,iy0,iz1),v(ix1,iy1,iz0),
				     v(ix1,iy1,iz1));
    }


  bool filled_bounds(int x, int y, int z, int xb, int yb, int zb)
    {
      return ( (x>=0) && (y>=0) && (z>=0) && 
	       (x<=xb) && (y<=yb) && (z<=zb) );
    }

  

  float filled_tri_interpolation(const volume& v, const float x, const float y,
			  const float z)
    {
      int ix0, iy0, iz0, ix1, iy1, iz1;
      float newx=x, newy=y, newz=z;
      ix0 = (int) floor(x);
      iy0 = (int) floor(y);
      iz0 = (int) floor(z);

      // extend the outer edges by one voxel

      if (ix0==v.xsize()-1) { ix0--; newx=v.xsize()-1.0; }
      else if (ix0==-1) { ix0++; newx=0.0; }
      // check if point is exactly an integer
      else if (x==(float) v.xsize()) { ix0=v.xsize()-2;  newx=v.xsize()-1.0; }
      else if (x==-1.0) { ix0=0;  newx=0.0; }

      if (iy0==v.ysize()-1) { iy0--; newy=v.ysize()-1.0; }
      else if (iy0==-1) { iy0++; newy=0.0; }
      // check if point is exactly an integer
      else if (y==(float) v.ysize()) { iy0=v.xsize()-2;  newy=v.ysize()-1.0; }
      else if (y==-1.0) { iy0=0;  newy=0.0; }

      if (iz0==v.zsize()-1) { iz0--; newz=v.zsize()-1.0; }
      else if (iz0==-1) { iz0++; newz=0.0; }
      // check if point is exactly an integer
      else if (z==(float) v.zsize()) { iz0=v.zsize()-2;  newz=v.zsize()-1.0; }
      else if (z==-1.0) { iz0=0;  newz=0.0; }

      ix1 = ix0 + 1;
      iy1 = iy0 + 1;
      iz1 = iz0 + 1;
      if ( (!v.in_bounds(ix0,iy0,iz0)) || (!v.in_bounds(ix1,iy1,iz1)) )
	{
	  cerr << "Tri_interpolation: Out of bounds!\n";
	  cerr << "(" << x << "," << y << "," << z << ") vs ";
	  cerr << "(" << v.columns() << "," << v.rows() << "," << 
	    v.slices() << ")" << endl;
	  return v.backgroundval();
	}
      return basic_tri_interpolation(v,newx-ix0,newy-iy0,newz-iz0,
				     v(ix0,iy0,iz0),
				     v(ix0,iy0,iz1),v(ix0,iy1,iz0),
				     v(ix0,iy1,iz1),v(ix1,iy0,iz0),
				     v(ix1,iy0,iz1),v(ix1,iy1,iz0),
				     v(ix1,iy1,iz1));
    }



  float bi_interpolation(const volume& v, const float x, const float y)
    {
      int ix0, iy0, ix1, iy1;
      ix0 = (int) floor(x);
      iy0 = (int) floor(y);
      ix1 = ix0 + 1;
      iy1 = iy0 + 1;
      if ((ix0==v.columns()-1) && ((x-(float)ix0)<1e-5))
	{
	  ix1=ix0;
	}
      if ((iy0==v.rows()-1) && ((y-(float)iy0)<1e-5))
	{
	  iy1=iy0;
	}
      if ( (!v.in_bounds(ix0,iy0,0)) || (!v.in_bounds(ix1,iy1,0)) )
	{
	  cerr << "Bi_interpolation: Out of bounds!\n";
	  cerr << "(" << x << "," << y << ") vs ";
	  cerr << "(" << v.columns() << "," << v.rows() << ")" << endl;
	  return 0.0;
	}
      return basic_tri_interpolation(v,x-ix0,y-iy0,0.0,v(ix0,iy0,0),
				     v(ix0,iy0,0),v(ix0,iy1,0),
				     v(ix0,iy1,0),v(ix1,iy0,0),
				     v(ix1,iy0,0),v(ix1,iy1,0),
				     v(ix1,iy1,0));
    }


#ifndef NO_NAMESPACE
 }
#endif
