/*  lazy.h

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

// Lazy evaluation support 


#if !defined(__lazy_h)
#define __lazy_h

#include <iostream>
#include <map>

#ifndef NO_NAMESPACE
 namespace LAZY {
#endif


typedef unsigned int uint;
typedef std::map<uint,bool,less<uint> > mapclass;
typedef std::map<uint,bool,less<uint> >::iterator mapiterator;


//-------------------------------------------------------------------------//

template <class T, class S>
class lazy {
private:
  mutable T storedval;
  uint tag;
  const S *iptr;
  T (*calc_fn)(const S &); 

private:
  T value() const;
  T calculate_val() const { return (*calc_fn)(*iptr); }

public:
  T operator() () const { return this->value(); }
  lazy() { tag = 0; }
  void init(const S *, T (*fnptr)(const S &));
  void copy (const lazy &, const S *);
};


//-------------------------------------------------------------------------//

class lazymanager {

template <class T, class S>  friend class lazy;

private:
  mutable bool validflag;
  mutable mapclass validcache;
  mutable uint tagnum;

private:
  uint getnewtag() const { return tagnum++; }

  const bool is_whole_cache_valid() const 
    { return validflag; }

  const bool is_cache_entry_valid(const uint tag) const 
    { return validcache[tag]; }
  void set_cache_entry_validity(const uint tag, const bool newflag) const 
    { validcache[tag] = newflag; }

  void invalidate_whole_cache() const;

public:
  lazymanager(); 
  void copylazymanager(const lazymanager &);
  void set_whole_cache_validity(const bool newflag) const 
    { validflag = newflag; }
};


//-------------------------------------------------------------------------//

// Body of lazy member functions (put here as cannot simply separate
//   templated definitions into seperate source files if building a library)



template <class T, class S>
T lazy<T,S>::value() const 
  {
    if ( (iptr == 0) || (tag==0) ) {
      cerr << "Error: uninitialized lazy evaluation class" << endl;
      exit(-1);
    }
    if (! iptr->is_whole_cache_valid() ) {
      iptr->invalidate_whole_cache();
      iptr->set_whole_cache_validity(true);
    }
    if (! iptr->is_cache_entry_valid(tag)) {
      //cerr << "Calculating value" << endl;
      storedval = calculate_val();
      iptr->set_cache_entry_validity(tag,true);
    }
    return storedval; 
  }
 

template <class T, class S>
void lazy<T,S>::init(const S *ip, T (*fnptr)(const S &)) 
  { 
    iptr = ip; 
    calc_fn = fnptr; 
    tag = iptr->getnewtag(); 
    iptr->set_cache_entry_validity(tag,false); 
  }
 


template <class T, class S>
void lazy<T,S>::copy (const lazy &source, const S *ip) {
  storedval = source.storedval;
  tag = source.tag;
  calc_fn = source.calc_fn; 
  // Do NOT copy the same parent class pointer 
  //   (allows parent class to be copied correctly)
  iptr = ip;
}


 
#ifndef NO_NAMESPACE
 }
#endif

#endif


