// Quick and dirty image storage class
//  Written by Mark Jenkinson  29/1/99

#include "mjimage.h"
#include "miscmaths.h"
#include <iostream>
#include <vector>
#include <algorithm>

#ifndef NO_NAMESPACE
 using namespace MISCMATHS;

 namespace MJIMAGE {
#endif

  basicvolume::basicvolume() : Data(0)
    {
      this->initialize(0,0,0,0);
    }


  basicvolume::basicvolume(const basicvolume& source) : Data(0)
    {
      this->initialize(source.columns(),source.rows(),source.slices(),0);
      this->copy(source);
    }

  basicvolume::basicvolume(int column, int row, int slice, float *d) : Data(d)
    {
      this->initialize(column,row,slice,d);
    }

  int basicvolume::initialize(int column, int row, int slice, float *d)
    {
      if (Data!=0)  delete [] Data;
      Xsize = 1.0;
      Ysize = 1.0;
      Zsize = 1.0;
      SlicesZ = slice;
      RowsY = row;
      ColumnsX = column;
      Bound = SlicesZ * RowsY * ColumnsX;
      if (Bound > 0) {
	if (d != 0)
	  Data = d;
	else
	  Data = new float[Bound];
      } else {
	Data = 0;
      }
      return 0;
    }


  int basicvolume::reinitialize(int column, int row, int slice, float *d)
    {
      return this->initialize(column,row,slice,d);
    }


  int basicvolume::reinitialize(const basicvolume& source)
    {
      if (source.Data==0) return -1;
      this->initialize(source.columns(),source.rows(),source.slices(),0);
      return this->copy(source);
    }



  basicvolume::~basicvolume()
    {  delete [] Data; }


  float basicvolume::set(float val)
    {
      for (unsigned int i=0; i<Bound; i++) {
	Data[i] = val;
      }
      return val;
    }


  float basicvolume::operator=(float val)
  {
    this->set(val);
    return val;
  }


  int basicvolume::copy(const basicvolume& source)
    {
      if (source.Data==0)  return -1;
      Xsize = source.Xsize;
      Ysize = source.Ysize;
      Zsize = source.Zsize;
      int NS=Min(SlicesZ,source.slices());
      int NR=Min(RowsY,source.rows());
      int NC=Min(ColumnsX,source.columns());
      for (int z=0; z<NS; z++) {
	for (int y=0; y<NR; y++) {
	  for (int x=0; x<NC; x++) {
	    this->operator()(x,y,z) = source(x,y,z);
	  }
	}
      }
      return 0;
    }
  

  /*
    void basicvolume::iterate_all(void (func)(float& ))
    {
    for (unsigned int i=0; i<Bound; i++) {
    func(Data[i]);
    }
    } 

 
    void basicvolume::iterate_all(void (func)(float& )) const
    {
    for (unsigned int i=0; i<Bound; i++) {
    func(Data[i]);
    }
    }  
  */


  ///////////////////////////////////////////////////////////////////////////
   

  volume::volume() 
    {
      v_sampling_matrix = identity(4);
      v_axis1.ReSize(3);
      v_planept1.ReSize(3);
      v_axis2.ReSize(3);
      v_planept2.ReSize(3);
      v_cog.ReSize(3);
      v_bval = -8888.0;
      this->invalidate_cog();
    }
  
  volume::volume(const volume& source) : basicvolume(source) 
    {
      v_sampling_matrix = source.v_sampling_matrix;
      this->setplanes(source.v_axis1,source.v_planept1,
		      source.v_axis2,source.v_planept2);
      v_cog = source.v_cog;
    }

  void volume::set_default_mat_and_cog() 
    {
      v_sampling_matrix = identity(4);
      v_bval = -8888.0;
      this->invalidate_cog();
    }

  void volume::set_default_planes() 
    {
      if (v_axis1.Nrows()!=3)     v_axis1.ReSize(3);
      if (v_planept1.Nrows()!=3)  v_planept1.ReSize(3);
      if (v_axis2.Nrows()!=3)     v_axis2.ReSize(3);
      if (v_planept2.Nrows()!=3)  v_planept2.ReSize(3);
      ColumnVector axis1(3), planept1(3), axis2(3), planept2(3);
      axis1 << 0.0 << 0.0 << 1.0;
      planept1 << 0.0 << 0.0 << -0.5;
      axis2 << 0.0 << 0.0 << -1.0;
      planept2 << 0.0 << 0.0 << (this->slices() - 0.5);
      this->setplanes(axis1,planept1,axis2,planept2);
    }

  volume::volume(int column, int row, int slice, float *d) :
    basicvolume(column,row,slice,d)
    {
      v_cog.ReSize(3);
      this->set_default_mat_and_cog();
      this->set_default_planes();
    }

  volume::~volume() 
    { 
      // do nothing
    }

  volume& volume::operator=(const volume& source)
  { 
    this->reinitialize(source);
    return *this;
  }
  
  int volume::reinitialize(int column, int row, int slice, float *d)
    {
      this->basicvolume::reinitialize(column,row,slice,d);
      this->set_default_mat_and_cog();
      this->set_default_planes();
      return 0;
    }

  int volume::reinitialize(const volume& source) 
    {
      v_sampling_matrix = source.v_sampling_matrix;
      v_cog = source.v_cog;
      v_bval = source.v_bval;
      this->setplanes(source.v_axis1,source.v_planept1,
		      source.v_axis2,source.v_planept2);
      this->basicvolume::operator=(source);
      this->setvoxelsize(this->getx(),this->gety(),this->getz());
      return 0;
    };
  
  Matrix volume::sampling_matrix() const
    {
      return v_sampling_matrix;
    }

  void volume::set_sampling_matrix(const Matrix& samplingmat) 
    {
      v_sampling_matrix = samplingmat;
    }

  //----------------------------------------------------------------------//

  void get_xintercepts(float& xmin, float& xmax, const ColumnVector& axis,
		       const ColumnVector& planept,
		       const unsigned int y, const unsigned int z,
		       const volume& v)
    {
      ColumnVector xhat(3), x1(3);
      xhat << 1.0 << 0.0 << 0.0;
      x1 << 0.0 << (Real) y << (Real) z;
      float xdota, epsilon = 1e-2, maxx, minx, lambda;
      maxx = v.columns() - 1.0;
      minx = 0.0;
      xdota = dot(xhat,axis);
      if (fabs(xdota)>epsilon) {
	lambda = dot(planept - x1,axis)/xdota;
	if (xdota > epsilon) 
	  { xmin = Max(lambda,minx);  xmax = maxx; }
	else 
	  { xmin = minx;  xmax = Min(lambda,maxx); }
      } else {
	xdota = dot(x1-planept,axis);
	if (xdota > 0.0)
	  { xmin = minx;  xmax = maxx; }
	else
	  { xmin = minx;  xmax = minx; }
      }
    }

  
  void get_xintercepts(unsigned int& x1, unsigned int& x2, 
		       const unsigned int y, const unsigned int z,
		       const volume& v)
    {
      ColumnVector ax1,pt1,ax2,pt2;
      v.getplanes(ax1,pt1,ax2,pt2);
      float x1a,x2a,x1b,x2b;
      get_xintercepts(x1a,x2a,ax1,pt1,y,z,v);
      get_xintercepts(x1b,x2b,ax2,pt2,y,z,v);
      int x1int, x2int;
      x1int = (int) (Max(x1a,x1b)+0.99);
      if (x1int>=v.columns())  x1int = v.columns() - 1;
      
      x2int = (int) Min(x2a,x2b);
      if (x2int<0)  x2int=0;
      if (x2int<x1int) {
	x1int = 0;
	x2int = 0;
      }
      x1 = (unsigned int) x1int;
      x2 = (unsigned int) x2int;
    }


  /*
  void get_xintercepts(unsigned int& x1, unsigned int& x2, 
		       const unsigned int y, const unsigned int z,
		       const volume& v1, const volume& v2)
    {
      unsigned int x1a, x1b, x2a, x2b;
      get_xintercepts(x1a,x2a,y,z,v1);
      get_xintercepts(x1b,x2b,y,z,v2);
      x1 = Max(x1a,x1b);
      x2 = Min(x2a,x2b);
      if (x2<x1) {
	x1 = 0;
	x2 = 0;
      }
    }
  */
  

  void volume::create_valid_mask(volume& mask) const
    {
      mask.reinitialize(*this);
      mask = 0.0;
      unsigned int zb = this->slices(), yb = this->rows(), x1, x2;
      for (unsigned int z=0; z<zb; z++) { 
	for (unsigned int y=0; y<yb; y++) { 
	  x1 = this->validxmin(y,z);
	  x2 = this->validxmax(y,z);
	  if (x1<0) {
	    cerr << "CREATE_VALID_MASK::validxmin " << x1 << " too small\n";
	    x1 = 0;
	  }
	  if ((x2>=(unsigned int) this->columns()) && (x1<=x2)) { 
	    cerr<<"CREATE_VALID_MASK::validxmax " << x2 << " too big\n";
	    x2=this->columns()-1;
	  }
	  for (unsigned int x=x1; x<=x2; x++) { 
	    mask(x,y,z) = 1.0;
	  }
	}
      }
    }


  void initialize_plane_bounds(basicvolume& valid, const volume& v)
    {
      float maxx = (float) v.columns() - 1.0,  minx = 0.0;
      unsigned int zb = v.slices(), yb = v.rows();
      for (unsigned int z=0; z<zb; z++) { 
	for (unsigned int y=0; y<yb; y++) { 
	  valid((unsigned int) 0,y,z) = minx-1.0;
	  valid((unsigned int) 1,y,z) = maxx+1.0;
	}
      }
    }


  
  void first_plane_bounds(basicvolume& valid, const ColumnVector& ax1,
			const ColumnVector& pt1, const volume& v)
    {
      // calculates the x bounds for the valid mask (first plane only)
      if ((ax1.Nrows()<3) || (pt1.Nrows()<3)) { 
	cerr << "First_plane_bounds::Invalid plane axis/point\n"; 
	return;
      }
      float maxx = (float) v.columns() - 1.0,  minx = 0.0;
      if ( fabs(ax1(1))<0.5/((float) Max(v.rows(),v.slices())) ) {
	unsigned int zb = v.slices(), yb = v.rows();
	float dotval, dot0 = -dot(pt1,ax1), axy = ax1(2), axz = ax1(3);
	for (unsigned int z=0; z<zb; z++) { 
	  dotval = dot0 + z*axz;
	  for (unsigned int y=0; y<yb; y++) { 
	    if (dotval<0.0) {
	      valid((unsigned int) 0,y,z) = minx-1.0;
	      valid((unsigned int) 1,y,z) = minx-1.0;
	    } else { // else keep the old bounds
	      valid((unsigned int) 0,y,z) = minx-1.0;
	      valid((unsigned int) 1,y,z) = maxx+1.0;
	    }
	    dotval += axy;
	  }
	}
	return;
      }

      float lambda, lambda0 = dot(pt1,ax1)/ax1(1);
      float ax_y_x = ax1(2)/ax1(1), ax_z_x = ax1(3)/ax1(1), axx=ax1(1);
      unsigned int zb = v.slices(), yb = v.rows();
      for (unsigned int z=0; z<zb; z++) { 
	lambda = lambda0 - z*ax_z_x;
	for (unsigned int y=0; y<yb; y++) { 
	  if (axx>0.0) {
	    valid((unsigned int) 0,y,z) = Max((float) (minx-1.0),lambda);
	    valid((unsigned int) 1,y,z) = maxx+1.0;
	  } else {
	    valid((unsigned int) 0,y,z) = minx-1.0;
	    valid((unsigned int) 1,y,z) = Min((float) (maxx+1.0),lambda);
	  }
	  lambda -= ax_y_x;
	}
      }
    }



  void add_plane_bounds(basicvolume& valid, const ColumnVector& ax1,
			const ColumnVector& pt1, const volume& v)
    {
      // calculates the x bounds for the valid mask (first plane only)
      if ((ax1.Nrows()<3) || (pt1.Nrows()<3)) { 
	cerr << "Add_plane_bounds::Invalid plane axis/point\n"; 
	return;
      }
      float minx = 0.0;
      if ( fabs(ax1(1))<0.5/((float) Max(v.rows(),v.slices())) ) {
	unsigned int zb = v.slices(), yb = v.rows();
	float dotval, dot0 = -dot(pt1,ax1), axy = ax1(2), axz = ax1(3);
	for (unsigned int z=0; z<zb; z++) { 
	  dotval = dot0 + z*axz;
	  for (unsigned int y=0; y<yb; y++) { 
	    if (dotval<0.0) {
	      valid((unsigned int) 0,y,z) = minx-1.0;
	      valid((unsigned int) 1,y,z) = minx-1.0;
	    } // else keep the old bounds
	    dotval += axy;
	  }
	}
	return;
      }

      float lambda, lambda0 = dot(pt1,ax1)/ax1(1);
      float ax_y_x = ax1(2)/ax1(1), ax_z_x = ax1(3)/ax1(1), axx=ax1(1);
      unsigned int zb = v.slices(), yb = v.rows();
      for (unsigned int z=0; z<zb; z++) { 
	lambda = lambda0 - z*ax_z_x;
	for (unsigned int y=0; y<yb; y++) { 
	  if (axx>0.0) {
	    valid((unsigned int) 0,y,z) 
	      = Max(valid((unsigned int) 0,y,z),lambda);
	    //valid(z,y,(unsigned int) 1) 
	    //  = valid(z,y,(unsigned int) 1);
	  } else {
	    //valid(z,y,(unsigned int) 0) 
	    //  = Max(valid(z,y,(unsigned int) 0),minx);
	    valid((unsigned int) 1,y,z) 
	      = Min(valid((unsigned int) 1,y,z),lambda);
	  }
	  lambda -= ax_y_x;
	}
      }
    }

  
  void fix_plane_bounds(basicvolume& valid, const volume& v)
    {
      float maxx = (float) v.columns() - 1.0,  minx = 0.0;
      unsigned int zb = v.slices(), yb = v.rows();
      for (unsigned int z=0; z<zb; z++) { 
	for (unsigned int y=0; y<yb; y++) { 
	  if (valid((unsigned int) 0,y,z)<minx) {
	    valid((unsigned int) 0,y,z) = minx;
	  }
	  if (valid((unsigned int) 1,y,z)>maxx) {
	    valid((unsigned int) 1,y,z) = maxx;
	  }
	  if ( (valid((unsigned int) 0,y,z)>maxx) 
	       || (valid((unsigned int) 1,y,z)<minx) ) {
	    valid((unsigned int) 0,y,z) = maxx+2.0;
	    valid((unsigned int) 1,y,z) = maxx+1.0;
	  } 
	}
      }
    }



  void volume::setplanes(const ColumnVector& axis1, 
			 const ColumnVector& planept1,
			 const ColumnVector& axis2,
			 const ColumnVector& planept2)
    {
      v_axis1 = axis1;
      v_planept1 = planept1;
      v_axis2 = axis2;
      v_planept2 = planept2;
      if ((v_validx.slices() != this->slices()) || 
	  (v_validx.rows() != this->rows()) || 
	  (v_validx.columns() != 2)) {
	v_validx.reinitialize(2,this->rows(),this->slices());
      }
      ColumnVector voxpt1(3), voxpt2(3), voxax1(3), voxax2(3);
      Matrix isamp(4,4);
      isamp = this->sampling_matrix().i();
      voxax1 = isamp.SubMatrix(1,3,1,3)*axis1;
      voxax2 = isamp.SubMatrix(1,3,1,3)*axis2;
      voxax1 = voxax1/norm2(voxax1);
      voxax2 = voxax2/norm2(voxax2);
      voxpt1 = isamp.SubMatrix(1,3,1,3)*planept1 + isamp.SubMatrix(1,3,4,4);
      voxpt2 = isamp.SubMatrix(1,3,1,3)*planept2 + isamp.SubMatrix(1,3,4,4);
      first_plane_bounds(v_validx,voxax1,voxpt1,*this);
      add_plane_bounds(v_validx,voxax2,voxpt2,*this);
      fix_plane_bounds(v_validx,*this);
    }


  void volume::getplanes(ColumnVector& axis1, ColumnVector& planept1,
			 ColumnVector& axis2, ColumnVector& planept2) const
    {
      axis1 = v_axis1;
      planept1 = v_planept1;
      axis2 = v_axis2;
      planept2 = v_planept2;
    }


  unsigned int volume::validxmin(const unsigned int y, const unsigned int z)
    const
    {
      return (unsigned int) v_validx((unsigned int) 0,y,z);
    }


  unsigned int volume::validxmax(const unsigned int y, const unsigned int z)
    const 
    {
      return (unsigned int) v_validx((unsigned int) 1,y,z);
    }
  

  //----------------------------------------------------------------------//

  int volume_get_cog(const volume& v, ColumnVector& cog)
    {
      cog(1)=0.0;
      cog(2)=0.0;
      cog(3)=0.0;
      float val=0.0, total=0.0, vmin=v(0,0,0);
      unsigned int zb = v.slices(), yb = v.rows(), xb = v.columns();
      for (unsigned int z=0; z<zb; z++) { 
	for (unsigned int y=0; y<yb; y++) { 
	  for (unsigned int x=0; x<xb; x++) { 
	    val = v(x,y,z);
	    cog(1) += val*x;
	    cog(2) += val*y;
	    cog(3) += val*z;
	    total += val;
	    if (val < vmin)   vmin = val;
	  }
	}
      }
      // correct for possible negative intensities
      total -= vmin * xb*yb*zb;
      cog(1) -= vmin * (xb-1)*xb*yb*zb/2;
      cog(2) -= vmin * (yb-1)*xb*yb*zb/2;
      cog(3) -= vmin * (zb-1)*xb*yb*zb/2;
      if (fabs(total) < 1e-5) {
	cerr << "WARNING::in calculating COG, total = 0.0" << endl;
	total = 1.0;
      }
      cog(1) /= total;
      cog(2) /= total;
      cog(3) /= total;
      // Now convert these values from voxel to physical space
      ColumnVector cog4(4);
      cog4.Rows(1,3) = cog.Rows(1,3);
      cog4(4) = 1.0;
      cog4 = v.sampling_matrix() * cog4;
      cog.Rows(1,3) = cog4.Rows(1,3);
      return 0;
    }


  bool is_valid_cog(const ColumnVector& cog)
    {
      ColumnVector invalid(3);
      invalid = -1.0;
      if (MaximumAbsoluteValue(cog - invalid)<1e-4) {
	return false;
      }
      return true;
    }

  void volume::invalidate_cog() const
    {
      this->v_cog = -1.0;
    }

  
  ColumnVector volume::calc_cog() const
    {
      volume_get_cog(*this,this->v_cog);
      if (!is_valid_cog(this->v_cog)) {
	this->v_cog+=1e-3;  // small adjustment to avoid invalid flag
      }
      return this->v_cog;
    }


  ColumnVector volume::cog() const
    {
      if (!is_valid_cog(this->v_cog)) {
	this->calc_cog();
      }
      return this->v_cog;
    }

  //-------------------------------------------------------------------//

  bool is_valid_bval(float bval) {
    return (fabs(bval+8888.0)>0.1);
  }
  
  
  float volume::calc_bval(unsigned int edgewidth) const
    {
      unsigned int zb = this->slices(), yb = this->rows(), xb = this->columns();
      unsigned int ew = edgewidth, numbins;
      numbins = 2*(xb-2*ew)*(yb-2*ew)*ew + 2*(xb-2*ew)*zb*ew + 2*yb*zb*ew;
      std::vector<float> hist(numbins);
      // put the edge voxel values into the histogram
      unsigned int hindx = 0;
      for (unsigned int e=0; e<ew; e++) {
	// put in the faces
	// xy faces (small lids of the box)
	for (unsigned int x=ew; x<xb-ew; x++) {
	  for (unsigned int y=ew; y<yb-ew; y++) {
	  hist[hindx++] = this->operator()(x,y,e);
	  hist[hindx++] = this->operator()(x,y,zb-1-e);
	  }
	}
	// xz faces (smallish edge faces)
	for (unsigned int x=ew; x<xb-ew; x++) {
	  for (unsigned int z=0; z<zb; z++) {
	  hist[hindx++] = this->operator()(x,e,z);
	  hist[hindx++] = this->operator()(x,yb-1-e,z);
	  }
	}
	// yz faces (large edge faces)
	for (unsigned int y=0; y<yb; y++) {
	  for (unsigned int z=0; z<zb; z++) {
	  hist[hindx++] = this->operator()(e,y,z);
	  hist[hindx++] = this->operator()(xb-1-e,y,z);
	  }
	}
      }
      sort(hist.begin(),hist.end());
      unsigned int percentile10 = numbins / 10;
      float v10 = hist[percentile10];
      if (!is_valid_bval(v10)) { v10 += 0.01*(hist[numbins-1] - hist[0]); }
      return v10;
    }


  float volume::backgroundval() const
    {
      if (!is_valid_bval(v_bval)) {
	v_bval = this->calc_bval(2);
      }
      return v_bval;
    }

  //---------------------------------------------------------------------//

  void volume::setvoxelsize(const float x, const float y, const float z)
    { 
      // set the diagonal components of the v_sampling_matrix to the voxelsize
      // AND make sure that if A0*X = 0 before, then A1*X = 0 after
      //  (that is, V = A0^-1 * [0 0 0 1]' is voxelorigin and
      //     for A1(3,3) * V = T1  =>  -T1 is the appropriate translation)
      ColumnVector t1(3), voxelorigin(4), worldorigin(4);
      worldorigin << 0.0 << 0.0 << 0.0 << 1.0;
      voxelorigin = v_sampling_matrix.i() * worldorigin;
      this->setx(x);  this->sety(y);  this->setz(z);
      v_sampling_matrix(1,1) = this->getx();
      v_sampling_matrix(2,2) = this->gety();
      v_sampling_matrix(3,3) = this->getz();
      t1 = v_sampling_matrix.SubMatrix(1,3,1,3) * 
	             voxelorigin.SubMatrix(1,3,1,1);
      v_sampling_matrix.SubMatrix(1,3,4,4) = -t1;
    }
  
  void volume::setvoxelorigin(const float x, const float y, const float z)
    {
      ColumnVector trans0(3), trans1(3);
      trans0 << x << y << z;
      trans1 = v_sampling_matrix.SubMatrix(1,3,1,3) * trans0;
      v_sampling_matrix.SubMatrix(1,3,4,4) = -trans1;
    }

  ColumnVector volume::getvoxelorigin() const
    {
      ColumnVector oret(3), voxelorigin(4), worldorigin(4);
      worldorigin << 0.0 << 0.0 << 0.0 << 1.0;
      voxelorigin = v_sampling_matrix.i() * worldorigin;
      oret = voxelorigin.SubMatrix(1,3,1,1);
      return oret;
    }

  tvolume::tvolume()
    {
      this->initialize(NULL, 0, 0);
    }
  
  tvolume::~tvolume() {}
  
  int tvolume::initialize(const volume* source, const int len, const float TR)
    {
      tr = TR;
      length = len;
      
      if (source == NULL){
	vols = NULL;
	return 0;
      }
      
      vols = new volume[len];
      
      for (int i=0; i < len; i++)
	vols[i] = source[i];
      
      return 0;
    }

  int tvolume::reinitialize(const volume* source, const int len, const float TR)
    {
      return this->initialize(source, len, TR);
    }

  
#ifndef NO_NAMESPACE
 }
#endif










